/*
 *            Vuze Vanish Backend GPL Version 2 License
 * 
 * The distributed software (Software) is named the "Vuze Vanish Backend" and
 * serves as a backend storage system for the Vanish system (described at
 * http://vanish.cs.washington.edu/). The Software consists of Vuze (a.k.a.
 * Azureus) Bittorrent client, version  4.2.0.2 (original and
 * latest source code of Vuze is available at http://azureus.sourceforge.net).
 * 
 * Copyright 2009, University of Washington (C) 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * All bug reports and technical questions shall be sent to:
 * vanish @ cs.washington.edu.
 * 
 */

package edu.washington.cs.vanish.internal.backend.impl;

import java.io.File;
import java.net.InetSocketAddress;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.apache.commons.configuration.ConfigurationException;
import org.gudy.azureus2.plugins.PluginInterface;

import com.aelitis.azureus.core.AzureusCore;
import com.aelitis.azureus.core.AzureusCoreFactory;
import com.aelitis.azureus.core.dht.DHT;
import com.aelitis.azureus.core.dht.DHTFactory;
import com.aelitis.azureus.core.dht.DHTLogger;
import com.aelitis.azureus.core.dht.DHTOperationAdapter;
import com.aelitis.azureus.core.dht.DHTStorageAdapter;
import com.aelitis.azureus.core.dht.control.DHTControl;
import com.aelitis.azureus.core.dht.impl.DHTLog;
import com.aelitis.azureus.core.dht.nat.DHTNATPuncherAdapter;
import com.aelitis.azureus.core.dht.transport.DHTTransportContact;
import com.aelitis.azureus.core.dht.transport.DHTTransportException;
import com.aelitis.azureus.core.dht.transport.DHTTransportValue;
import com.aelitis.azureus.core.dht.transport.udp.DHTTransportUDP;
import com.aelitis.azureus.core.dht.transport.udp.impl.DHTTransportUDPImpl;
import com.aelitis.azureus.plugins.dht.impl.DHTPluginStorageManager;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.conf.VanishConfiguration.TimeUnit;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.ShareValidator;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.TooManySharesLostVanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.backend.VanishBackendException;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;

/**
 * Vuze-based implementation of the VanishBackendInterface.
 * 
 * @author roxana
 */

class VuzeDHTParams implements RefreshableParams {	
	final int  kDhtIndexLen;
	final int  kNumTries;
	final long kRetrySleep_ms;
	
	final int  kNumTriesIntegrate;
	final long kIntegrateRetryInterval_ms;
	
	final int kDdhtOperationBulkSize;
	final long kTimeoutBetweenOperationBulks;
	
	public VuzeDHTParams(int dhtIndexLen, int numTries, int retrySleep_ms,
			int dhtOperationBulkSize, long timeoutBetweenOperationBulks,
			int numTriesIntegrate, long integrateRetryInterval_ms) {
		this.kDhtIndexLen   = dhtIndexLen;
		this.kNumTries      = numTries;
		this.kRetrySleep_ms = retrySleep_ms;
		this.kDdhtOperationBulkSize = dhtOperationBulkSize;
		this.kTimeoutBetweenOperationBulks = timeoutBetweenOperationBulks;
		this.kNumTriesIntegrate = numTriesIntegrate;
		this.kIntegrateRetryInterval_ms = integrateRetryInterval_ms;
	}
}

public class VuzeVanishBackendImpl
implements DHTNATPuncherAdapter, VanishBackendInterface,
           RefreshableConfig<VuzeDHTParams>, Constants {	
	// DHT-related params (these are fixed forever, as we give them to DHT):
	private final boolean kDhtLoggingOn;
	private final int     kDhtLookupConcurrency;

	private final int     kDhtUDPOperationTimeoutMin_ms;
	private final int     kDhtUDPOperationTimeoutMax_ms;
	private final int     kDhtUDPTimeoutSafeDistanceFromMin_ms;
	private final long    kDhtUDPTimeoutIntervalRecalibrationPeriod;
	
	private final String  kDhtBootstrapAddress;
	private final int     kDhtBootstrapPort;
	
	private final int     kDhtNetwork;
	private final byte     kDhtProtocolVersion;

	private int           kDhtPort;
	private final int     kDhtNumReplicas;
	private final String  kDhtDirectory;
	
	private final DHTLogger kDhtLogger;
	private final DHTStorageAdapter kStorageManager;

	private final Properties kDhtProperties;
	
	// Vanish backend params, which are reloadable from the Configuration.
	private VuzeDHTParams params;

	// State:

	private final VanishLogger LOG;
	private final VanishConfiguration CONF;
	
	private final Timer timeout_recalibration_process;
	
	/** The DHT object. */
	private DHT dht = null;  // null until init is called
	private ConfigurableTimeoutDHTTransport transport = null;
	/** Protects the DHT and transport. */
	ReentrantReadWriteLock dht_lock = new ReentrantReadWriteLock();
	
	private int current_udp_timeout_range_min;
	private int current_udp_timeout_range_max;
	private int current_udp_timeout;

	public VuzeVanishBackendImpl(VanishConfiguration config,
			                     VanishLogger LOG)
	throws InvalidArgumentVanishException {
		this.LOG = LOG;
		this.CONF = loadConfig(config);
		
		// Load the parameters from the configuration:

		kDhtLoggingOn = CONF.getBoolean(
				VuzeBackendDefaults.CONF_VUZE_DHT_LOGGING_ON,
				VuzeBackendDefaults.CONF_VUZE_DHT_LOGGING_ON_DEFAULT);
		kDhtLookupConcurrency = new Integer(CONF.getInt(
				VuzeBackendDefaults.CONF_VUZE_LOOKUP_CONCURRENCY,
				VuzeBackendDefaults.CONF_VUZE_LOOKUP_CONCURRENCY_DEFAULT));
		kDhtPort = CONF.getInt(
				VuzeBackendDefaults.CONF_VUZE_PORT,
				VuzeBackendDefaults.CONF_VUZE_PORT_DEFAULT);
		kDhtNumReplicas = CONF.getInt(
				VuzeBackendDefaults.CONF_VUZE_REPLICATION_FACTOR,
				VuzeBackendDefaults.CONF_VUZE_REPLICATION_FACTOR_DEFAULT);

		kDhtUDPOperationTimeoutMin_ms = CONF.getInt(
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_MIN,
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_MIN_DEFAULT);
		kDhtUDPOperationTimeoutMax_ms = CONF.getInt(
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_MAX,
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_MAX_DEFAULT);
		kDhtUDPTimeoutSafeDistanceFromMin_ms = CONF.getInt(
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_SAFE_DISTANCE_FROM_MIN,
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_SAFE_DISTANCE_FROM_MIN_DEFAULT);
		kDhtUDPTimeoutIntervalRecalibrationPeriod = CONF.getTimeInMillis(
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_RECALIBRATE_INTERVAL,
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_RECALIBRATE_INTERVAL_DEFAULT,
				VanishConfiguration.TimeUnit.HOURS);
		
		String full_addr = CONF.getString(
				VuzeBackendDefaults.CONF_VUZE_BOOTSTRAP_NODE,
				VuzeBackendDefaults.CONF_VUZE_BOOTSTRAP_NODE_DEFAULT);
		kDhtBootstrapAddress = getHostname(full_addr);
		kDhtBootstrapPort = getPort(full_addr);
		
		kDhtNetwork = full_addr.equals(VuzeBackendDefaults.AELITIS_ADDRESS)
				? DHT.NW_MAIN : DHT.NW_CVS;
		kDhtProtocolVersion = (kDhtNetwork == DHT.NW_MAIN 
				? DHTTransportUDP.PROTOCOL_VERSION_MAIN
				: DHTTransportUDP.PROTOCOL_VERSION_CVS);

		kDhtDirectory = CONF.getString(
				VuzeBackendDefaults.CONF_VUZE_STORAGE_DIR,
				VuzeBackendDefaults.CONF_VUZE_STORAGE_DIR_DEFAULT);
		
		kDhtLogger = getLogger();
		kStorageManager = createStorageAdapter(
				kDhtNetwork, kDhtLogger, new File(kDhtDirectory));
		
		kDhtProperties = constructDHTProperties();
		
		params = this.getRenewedParamsFromConfig(CONF);
		setParams(params);

		timeout_recalibration_process = new Timer("timeoutrecalib", true);
		
        // Initialize non-final state:
		initState();
	}
	
	private VanishConfiguration loadConfig(VanishConfiguration config) {
		VanishConfiguration new_config = config;
		try {
			new_config = new VanishConfiguration(
					VuzeBackendDefaults.VUZE_BACKEND_CONFIGURATION_FILE);
		} catch (ConfigurationException e) {  // Ignore errors.
		}
		return new_config;
	}
	
	private void initState() throws InvalidArgumentVanishException {
		resetTimeoutInterval();
		
		if (current_udp_timeout_range_max < current_udp_timeout_range_min) {
			throw new InvalidArgumentVanishException(
					"Invalid UDP timeout range");
		}
		
		current_udp_timeout = (current_udp_timeout_range_min +
	               current_udp_timeout_range_max) / 2;
	}
	
	/**
	 * @param full_address  machine:port format.
	 * @return
	 * @throws VanishBackendException 
	 */
	private String getHostname(String full_address)
	throws InvalidArgumentVanishException {
		String ret = full_address;
		try {
			ret = full_address.substring(0, full_address.indexOf(':'));
		} catch (IndexOutOfBoundsException e) {
			throw new InvalidArgumentVanishException(
					"Invalid bootstrap address.");
		}
		return ret;
	}
	
	private int getPort(String full_address)
	throws InvalidArgumentVanishException {
		int ret = 0;
		try {
			ret = Integer.parseInt(full_address.substring(
					full_address.indexOf(':') + 1));
		} catch (Exception e) {
			throw new InvalidArgumentVanishException(
					"Invalid bootstrap address.");
		}
		return ret;
	}
	
    // Initializable interface:

	/**
	 * Thread-safe.
	 */
	// @Override
	public void init() throws VanishBackendException {
		final LogStat log_stat = LOG.logStat("DHTVanishBackend.init");
		
		dht_lock.writeLock().lock();
		if (isInitialized()) throw new VanishBackendException("Already init");
		try {
			startDHTNode(current_udp_timeout);
		} finally {
			dht_lock.writeLock().unlock();
		}
		
		if (kDhtUDPTimeoutIntervalRecalibrationPeriod > 0) {
			timeout_recalibration_process.schedule(
					new TimeoutRecalibrationTask(),
					kDhtUDPTimeoutIntervalRecalibrationPeriod,
					kDhtUDPTimeoutIntervalRecalibrationPeriod);
		}
		log_stat.end();
	}

	/**
	 * Not thread-safe!
	 */
	// @Override
	public boolean isInitialized() {
		boolean ret = false;
		ret = (dht != null);
		return ret;
	}

	// @Override
	public void stop() throws VanishException {
		if (dht != null) dht.destroy();
		AzureusCore core = AzureusCoreFactory.getSingleton();
		core.stop();
	}
	
	// Helper functions:
	
	private Properties constructDHTProperties() {
		DHTTransportUDPImpl.TEST_EXTERNAL_IP = false;  // TODO(roxana): true?
		
		Properties dht_props = new Properties();
		dht_props.put(DHT.PR_CONTACTS_PER_NODE, kDhtNumReplicas);
		dht_props.put(DHT.PR_LOOKUP_CONCURRENCY, kDhtLookupConcurrency);
		dht_props.put(DHT.PR_SEARCH_CONCURRENCY, kDhtLookupConcurrency);
		dht_props.put(DHT.PR_CACHE_REPUBLISH_INTERVAL,
				new Integer(DHTControl.CACHE_REPUBLISH_INTERVAL_DEFAULT));

		DHTLog.logging_on = kDhtLoggingOn;
		
		return dht_props;
	}
	
	private class ConfigurableTimeoutDHTTransport extends DHTTransportUDPImpl {
		public ConfigurableTimeoutDHTTransport(
				byte                    protocol_version,
                int                     network,
                boolean                 v6,
                String                  ip,
                String                  default_ip,
                int                             port,
                int                             max_fails_for_live,
                int                             max_fails_for_unknown,
                long                    timeout,
                int                             send_delay,
                int                             receive_delay,
                boolean                 bootstrap_node,
                boolean                 reachable,
                DHTLogger               logger )
		throws DHTTransportException {
			super(  protocol_version,
                    network,
                    v6,
                    ip,
                    default_ip,
                    port,
                    max_fails_for_live,
                    max_fails_for_unknown,
                    timeout,
                    send_delay,
                    receive_delay,
                    bootstrap_node,
                    reachable,
                    logger );
		}
		
		/**
		 * Override to make it visible here.
		 */
		@Override
		protected void setTimeout(long timeout)
		throws DHTTransportException {
			 super.setTimeout(timeout);
		}
	}
	
	/**
	 * Must be called while holding the writelock on dht_lock.
	 * @param bootstrapAddress
	 * @throws VanishBackendException
	 */
	protected void startDHTNode(int udp_timeout)
	throws VanishBackendException {
		try{
			transport = new ConfigurableTimeoutDHTTransport(
					kDhtProtocolVersion, kDhtNetwork, false, null, null,
					kDhtPort, 3, 1, udp_timeout, 25, 25, false, false,
					kDhtLogger);
			dht = DHTFactory.create(transport, kDhtProperties,
					kStorageManager, this, kDhtLogger);
			Throwable exception = null;
			for (int i = 0; i < params.kNumTriesIntegrate; ++i) {
				try {
					((DHTTransportUDP)transport).importContact(
							new InetSocketAddress(kDhtBootstrapAddress,
									              kDhtBootstrapPort),
							kDhtProtocolVersion);
					dht.integrate(true);
					exception = null;
					break;  // integration completed successfully.
				} catch (Throwable e) {
					exception = e;
					Thread.sleep(params.kIntegrateRetryInterval_ms);
				}
			}
			if (exception != null) throw exception;
		} catch(Throwable e) {
			throw new VanishBackendException("Failed to create dht", e);
		}
	}

	protected DHTStorageAdapter createStorageAdapter(int network,
			                                         DHTLogger logger,
			                                         File storage_dir) {
		return new DHTPluginStorageManager(network, logger, storage_dir);
	}
	
	protected DHTLogger getLogger() {		
		final PluginInterface plugin_interface = AzureusCoreFactory.create()
				.getPluginManager().getDefaultPluginInterface();
		
		DHTLogger ret_logger = new DHTLogger() {
				public void log(String str) {
					if (DHTLog.logging_on) LOG.debug(str);
				}
			
				public void log(Throwable e) {
					if (DHTLog.logging_on) LOG.error("", e);
				}
				
				public void log(int log_type, String str) {
					if (isEnabled(log_type)) LOG.debug(str);
				}

				public boolean isEnabled(int log_type) {
					return DHTLog.logging_on;
				}
					
				public PluginInterface getPluginInterface() {
					return plugin_interface;
				}
			};

		return ret_logger;
	}
	
	// VanishBackend interface:

	// @Override
	public byte[][] getShares(byte[][] locations, int needed_keys)
	throws VanishBackendException, TooManySharesLostVanishException {
		if (! isInitialized()) throw new VanishBackendException("uninit");
		final LogStat log_stat = LOG.logStat("DHTVanishBackend.getShares");

		byte[][] results = null;
		dht_lock.readLock().lock();
		try {
			BulkOperationResults results_aggregator =
				new GetSharesBulkOperationResults(locations.length,
						needed_keys, null);
			performOperation(
				locations,
				null,
				new BulkOperation() {
					public void initiate(byte[] index,
							@SuppressWarnings("unused") byte[] value,
							DHTOperationAdapter adapter)
					throws VanishBackendException {
						get(index, adapter);
					}
				},
				results_aggregator);
		
			results = results_aggregator.getResultsSnapshot();
			if (! results_aggregator.gotEnoughValidResponses()) {
				int num_valid_results = 0;
				for (int i = 0; i < results.length; ++i) {
					if (results[i] != null) ++num_valid_results;
				}
				// Oops, lost too many keys...
				throw new TooManySharesLostVanishException("Got " +
					num_valid_results + "/" + needed_keys + " shares");
			}
		} finally {
			dht_lock.readLock().unlock();
		}

		log_stat.end();
		return results;
	}

	// @Override
	public void pushShares(byte[][] keys, byte[][] locations)
	throws VanishBackendException {
		if (! isInitialized()) throw new VanishBackendException("Not init");
		final LogStat log_stat = LOG.logStat("DHTVanishBackend.pushShares");
		
		dht_lock.readLock().lock();
		try {
			BulkOperationResults results_aggregator =
				new PutSharesBulkOperationResults(keys.length);
        
			performOperation(
					locations,
					keys,
					new BulkOperation() {
						public void initiate(byte[] index,
								@SuppressWarnings("unused") byte[] value,
								DHTOperationAdapter adapter)
						throws VanishBackendException {
							put(index, value, adapter);
						}
					},
					results_aggregator);
		
			if (! results_aggregator.gotEnoughValidResponses()) {  // Failure.
				reportFailure();
				throw new VanishBackendException("Failed to push all shares.");
			} else {  // Success.
				reportSuccess();
			}
		} finally {
			dht_lock.readLock().unlock();
		}

		log_stat.end();
	}
	
	/**
	 * Must be called while holding readlock on dht_lock.
	 */
	private void reportFailure() throws VanishBackendException {
		if (current_udp_timeout_range_max - current_udp_timeout > 0) {
			reconfigureDHTTimeout(false);
		}
	}
	
	/**
	 * Must be called while holding readlock on dht_lock.
	 */
	private void reportSuccess() throws VanishBackendException {
		if (current_udp_timeout - current_udp_timeout_range_min >
		    kDhtUDPTimeoutSafeDistanceFromMin_ms) {
			reconfigureDHTTimeout(true);
		}
	}
	
	protected int getNewTimeout(boolean success) {
		if (success) {  // Decrease
			current_udp_timeout_range_max = current_udp_timeout;
		} else {  // Increase
			current_udp_timeout_range_min = current_udp_timeout;
			if (current_udp_timeout_range_max - current_udp_timeout_range_min
				<= 1) {
				current_udp_timeout_range_max = kDhtUDPOperationTimeoutMax_ms;
			}
		}
		int new_udp_timeout = (current_udp_timeout_range_min +
                current_udp_timeout_range_max) / 2;
		return new_udp_timeout;
	}
	
	/**
	 * Must be called while holding the readlock on dht_lock.
	 * Takes the writelock on dht_lock.
	 * 
	 * Pretty expensive operation, as we take writelock.
	 * 
	 * @throws VanishBackendException 
	 */
	private void reconfigureDHTTimeout(boolean success)
	throws VanishBackendException {
		dht_lock.readLock().unlock();
		dht_lock.writeLock().lock();
		
		int new_udp_timeout = getNewTimeout(success);
		
		try {
			if (current_udp_timeout == new_udp_timeout) return;
			
			current_udp_timeout = new_udp_timeout;
			
			LOG.info("DHTVanishBackend: adjusting UDP timeout to " +
				     new_udp_timeout);
			if (transport != null) {
				try {
					transport.setTimeout(current_udp_timeout);
				} catch (DHTTransportException e) {
					throw new VanishBackendException(
							"Failed to update timeout", e);
				}
			}
		} finally {
			dht_lock.readLock().lock();
			dht_lock.writeLock().unlock();
		}
	}
	
	/**
	 * CAlled only while holding writerlock on dht_lock.
	 */
	private void resetTimeoutInterval() {
		current_udp_timeout_range_max = kDhtUDPOperationTimeoutMax_ms;
		current_udp_timeout_range_min = kDhtUDPOperationTimeoutMin_ms;
	}

	private class TimeoutRecalibrationTask extends TimerTask {
		@Override
		public void run() {
			LOG.info("DHTVanishBackend#TimeoutRecalibrationTask reenabling " +
					"recalibration");
			dht_lock.writeLock().lock();
			resetTimeoutInterval();
			dht_lock.writeLock().unlock();
		}
	}

	// @Override
	public byte[] generateLocation(Random prg) {
		byte[] index = new byte[params.kDhtIndexLen];
		prg.nextBytes(index);
		return index;
	}

	// @Override
	public byte[] getID() {
		byte[] res = null;
		dht_lock.readLock().lock();
		res = dht.getControl().getRouter().getLocalContact().getID();
		dht_lock.readLock().unlock();
		return res;
	}
	
	// Helpers:
	
	private interface BulkOperation {
		public void initiate(byte[] index, byte[] value,
				             DHTOperationAdapter adapter)
		throws VanishBackendException;
	}
	
	private abstract class BulkOperationResults {
		private byte[][] results;
		private int num_expected_valid_results;
		
		private int num_valid_results_obtained_so_far;
		private int num_operations_completed_so_far;
		
		public BulkOperationResults(int num_total_possible_results,
				                    int num_expected_valid_results) {
			this.results = new byte[num_total_possible_results][];
			for (int i = 0; i < results.length; ++i) results[i] = null;
			
			this.num_expected_valid_results = num_expected_valid_results;
			
			num_valid_results_obtained_so_far = 0;
			num_operations_completed_so_far   = 0;
		}
		
		public void addRespose(int index, byte[] resp) {
			if (isValidResponse(index, resp)) {
				synchronized(this) {
					if (! alreadyHasValidResponse(index)) {
						results[index] = resp;
						++num_valid_results_obtained_so_far;
					}
					if (gotEnoughValidResponses()) this.notify();
				}
			}
		}
		
		public void setComplete(int index) {
			synchronized(this) {
				++num_operations_completed_so_far;
				if (completedAllOperations()) this.notify();
			}
		}
		
		protected abstract boolean isValidResponse(int index, byte[] resp);
		
		public synchronized boolean alreadyHasValidResponse(int index) {
			return results[index] != null;
		}
		
		public synchronized boolean gotEnoughValidResponses() {
			return num_valid_results_obtained_so_far >=
				   num_expected_valid_results;
		}
		
		public synchronized boolean completedAllOperations() {
			return num_operations_completed_so_far >= results.length;
		}
		
		protected byte[][] getResults() { return results; }
		
		public synchronized byte[][] getResultsSnapshot() {
			return Arrays.asList(results).toArray(new byte[0][]);
		}
	}
	
	private class GetSharesBulkOperationResults
	extends BulkOperationResults {
		ShareValidator validator;
		public GetSharesBulkOperationResults(int num_total_possible_results,
				                         int num_expected_valid_results,
				                         ShareValidator validator) {
			super(num_total_possible_results, num_expected_valid_results);
			this.validator = validator;
		}

		@Override
		protected boolean isValidResponse(int index, byte[] resp) {
			if (resp == null) {
				return false;
			}
			else {
				if (validator != null) return validator.isValid(index, resp);
				else return true;
			}
		}
	}
	
	private class PutSharesBulkOperationResults
	extends BulkOperationResults {
		public PutSharesBulkOperationResults(int num_total_values_pushed) {
			super(num_total_values_pushed, num_total_values_pushed);
		}

		@Override
		protected boolean isValidResponse(int index, byte[] resp) {
			return resp != null;
		}
	}
	
	private class DHTListener extends DHTOperationAdapter {		
		int identifier;
		private BulkOperationResults response_cb;
		private byte[] delayed_result;
				
		public DHTListener(Integer identifier,
				           BulkOperationResults response_cb) {
			this.identifier = identifier;
			this.response_cb = response_cb;
		}
		
	    public void wrote(DHTTransportContact contact, 
	   		 		      DHTTransportValue value) {
	    	byte[] v = value.getValue();
	    	if (LOG.isDebugEnabled()) {
	    		LOG.debug("Put: result: " + identifier + DHTLog.getString(v) +
	    				  " to contact " + contact.getAddress());
	    	}
	    	delayed_result = v;  // pointer modification is atomic.
	    }
	    
	    public void read(DHTTransportContact contact,
	    		 		              DHTTransportValue value) {
	    	byte[] v = value.getValue();
	    	if (LOG.isDebugEnabled()) {
	    		LOG.debug("Get: result: " + identifier + DHTLog.getString(v) + 
	          		      " from contact " + contact.getAddress());
	    	}
	    	response_cb.addRespose(identifier, v);
	    }
	    
	    public void complete(boolean timeout) {
	    	 if (LOG.isDebugEnabled()) {
	    		 if (timeout) LOG.debug("Get/put timed out: "+ identifier);
	    		 else LOG.debug("Get/put: COMPLETE: " + identifier);
	    	 }
	    	 
	    	 // TODO(roxana):
	    	 // Do I have to worry about wrote/read - complete synchronization
	    	 // issues? Don't think so, as complete shoudl be the last
	    	 // operation (i.e., no wrote/read can appear during/after it,
	    	 // nor can it appear during a wrote/read).
	    	 
	    	 if (!timeout && delayed_result != null) {
	    		 response_cb.addRespose(identifier, delayed_result);
	    	 }
	    	 response_cb.setComplete(identifier);
	    }
	}
	
	private void performOperation(byte[][] locations, byte[][] payloads,
			BulkOperation operation,
			BulkOperationResults results_aggregator)
	throws VanishBackendException {
		for (int trial_no = 0; trial_no < params.kNumTries; ++trial_no) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("DHTVanishBackend.performOperation trial " +
					      (trial_no + 1) + " start");
			}
			// Operation initiate phase:
			int num_ops_issued = 0;
			for (int index = 0; index < locations.length; ++index) {
				if (! results_aggregator.alreadyHasValidResponse(index)) {
					// Don't yet have a valid response for this index.
					DHTOperationAdapter dhtlistener = new DHTListener(index,
							results_aggregator);
					operation.initiate(
							locations[index],
							(payloads == null ? null : payloads[index]),
							dhtlistener);

					++num_ops_issued;

					if (num_ops_issued % params.kDdhtOperationBulkSize == 0) {
						// Wait in between consec. bulks, s.t. we don't
						// overload the transport and have ops be dropped.
						try {
							Thread.sleep(params.kTimeoutBetweenOperationBulks);
						} catch (InterruptedException e) { }
					} else {  // just yield
						Thread.yield();
					}
				}
			}
			
			// Response wait phase:
			long timeout_time = System.currentTimeMillis() +
			                    getWaitTimeForNumOperations(num_ops_issued);
			long left = 0;
			synchronized(results_aggregator) {
				while (! results_aggregator.gotEnoughValidResponses() &&
					   ! results_aggregator.completedAllOperations() &&
					   (left = timeout_time - System.currentTimeMillis()) > 0) {
					try {
						results_aggregator.wait(left);
						break;
					} catch (InterruptedException e) {
						continue;
					}
				}
				
				if (results_aggregator.gotEnoughValidResponses()) break;
			}
			
			if (params.kRetrySleep_ms != 0) {
				try { Thread.sleep(params.kRetrySleep_ms); }
				catch (InterruptedException e) { }
			}
			
			if (LOG.isDebugEnabled()) {
				LOG.debug("DHTVanishBackend.performOperation trial " +
					  (trial_no + 1) + " end");
			}
		}
	}
	
	private long getWaitTimeForNumOperations(int num_operations) {
		return num_operations * current_udp_timeout * 20;
	}
	
	
	// Protected to enable testing.
	protected void put(byte[] key, byte[] value, DHTOperationAdapter adapter) {
		dht.put(key, "", value, (byte)0, adapter);
	}
	
	// Protected to enable testing.
	protected void get(byte[] key, DHTOperationAdapter adapter) {
		dht.get(key, "", (byte)0, 32, 60000/*getWaitTimeForNumOperations(1)*/, false,
				false, adapter);
	}
	
	// DHTNATPuncherInterface:

	// @Override
	@SuppressWarnings("unchecked")
	public Map getClientData(InetSocketAddress originator,
			                 Map originator_data) {
		Map	res = new HashMap();
		res.put("udp_data_port", this.dht.getTransport().getPort());
		res.put("tcp_data_port", this.dht.getTransport().getPort());
		return res;
	}

	// RefreshableConfig interface:
	
	// @Override
	public VuzeDHTParams getRenewedParamsFromConfig(VanishConfiguration config) {		
		final int kDhtIndexLen = config.getInt(
				VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_DHTINDEXLEN,
				VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_DHTINDEXLEN_DEFAULT);
		final int kNumTries = config.getInt(
				VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_NUMTRIES,
				VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_NUMTRIES_DEFAULT);
		final int kRetrySleep_ms = config.getInt(
				VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_RETRYSLEEP_MS,
				VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_RETRYSLEEP_MS_DEFAULT);
		/*final int kTimeoutWaitForShare_ms = config.getLong(
				Defaults.CONF_VANISH_BACKEND_VUZEIMPL_SHARETIMEOUT_MS,
				10 * kDhtUDPOperationTimeoutMax_ms);*/
		
		final int kOperationBulkSize = config.getInt(
			VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_OPERATION_BULKSIZE,
			VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_OPERATION_BULKSIZE_DEFAULT);
		
		final long kTimeoutBetweenOperationBulks = config.getLong(
		  VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_DELAY_OPERATION_BULKS,
		  VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_DELAY_OPERATION_BULKS_DEFAULT);
		
		final int kNumTriesIntegrate = config.getInt(
	      VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_INTEGRATION_TRIALS,
		  VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_INTEGRATION_TRIALS_DEFAULT);
		
		final long kIntegrateInterval_ms = config.getTimeInMillis(
		  VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_INTEGRATION_TRIAL_INTERVAL_S,
		  VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_INTEGRATION_TRIAL_INTERVAL_S_DEFAULT,
		  TimeUnit.SECONDS);
		
		VuzeDHTParams params = new VuzeDHTParams(
				kDhtIndexLen,
				kNumTries, kRetrySleep_ms,
				kOperationBulkSize, kTimeoutBetweenOperationBulks,
				kNumTriesIntegrate, kIntegrateInterval_ms);
		
		// validate(params);
		return params;
	}

	// @Override
	public void setParams(VuzeDHTParams params) {
		this.params = params;
	}

	public String getIdentifier() {
		return "vuze";  // conflicts w/ the RPC one.
	}
}
