/*
 *            Vuze Vanish Backend GPL Version 2 License
 * 
 * The distributed software (Software) is named the "Vuze Vanish Backend" and
 * serves as a backend storage system for the Vanish system (described at
 * http://vanish.cs.washington.edu/). The Software consists of Vuze (a.k.a.
 * Azureus) Bittorrent client, version  4.2.0.2 (original and
 * latest source code of Vuze is available at http://azureus.sourceforge.net).
 * 
 * Copyright 2009, University of Washington (C) 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * All bug reports and technical questions shall be sent to:
 * vanish @ cs.washington.edu.
 * 
 */

package edu.washington.cs.vanish.internal.backend.impl;

import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import org.gudy.azureus2.core3.util.HashWrapper;

import com.aelitis.azureus.core.dht.DHTLogger;
import com.aelitis.azureus.core.dht.DHTOperationAdapter;
import com.aelitis.azureus.core.dht.DHTStorageAdapter;
import com.aelitis.azureus.core.dht.netcoords.DHTNetworkPosition;
import com.aelitis.azureus.core.dht.transport.DHTTransport;
import com.aelitis.azureus.core.dht.transport.DHTTransportContact;
import com.aelitis.azureus.core.dht.transport.DHTTransportException;
import com.aelitis.azureus.core.dht.transport.DHTTransportFullStats;
import com.aelitis.azureus.core.dht.transport.DHTTransportReplyHandler;
import com.aelitis.azureus.core.dht.transport.DHTTransportValue;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.TooManySharesLostVanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendException;
import edu.washington.cs.vanish.internal.backend.impl.VuzeVanishBackendImpl;
import edu.washington.cs.vanish.logging.VanishLogger;

import junit.framework.TestCase;

/**
 * 
 * @author roxana
 *
 */
class TestDHTVanishBackend extends VuzeVanishBackendImpl {	
	protected Map<HashWrapper,DHTTransportValue> storage;
	private boolean is_init = false;
	
	private static final TestDHTTransportContact contact =
		new TestDHTTransportContact(0);
	
	Timer send_answers_timer;
	
	// Timing parameters.
	boolean put_times_out_before_completion = false;
	
	long operation_delay = 10;  // in ms
	
	private int operation_count = 0;
	
	int count_of_operation_with_additional_delay = -1;  // none
	long additional_delay = 0;
	
	public TestDHTVanishBackend(VanishConfiguration conf)
	throws InvalidArgumentVanishException {
		super(conf, VanishLogger.getLogger("test"));
		this.storage = Collections.synchronizedMap(
				new HashMap<HashWrapper, DHTTransportValue>());
		send_answers_timer = new Timer();
	}
	
	protected void reset() {
		this.storage.clear();
		this.operation_count = 0;
		this.count_of_operation_with_additional_delay = -1;
		this.additional_delay = 0L;
		this.send_answers_timer.purge();
	}
	
	@Override
	public boolean isInitialized() { return is_init; }
	
	@Override
	public void stop() { }
	
	@Override
	protected void startDHTNode(int udp_timeout)
	throws VanishBackendException { is_init = true; }
	
	@Override
	protected DHTStorageAdapter createStorageAdapter(int network,
            DHTLogger logger,
            File storage_dir) {
		return null;
	}
	
	@Override
	protected DHTLogger getLogger() {
		return null;
	}

	@Override
	protected void put(final byte[] key_bytes, final byte[] value_bytes,
			final DHTOperationAdapter adapter) {
		++operation_count;
		
		final DHTTransportValue value = new TestDHTTransportValue(
				contact, value_bytes, false);
		final HashWrapper key = new HashWrapper(key_bytes);
		send_answers_timer.schedule(
				new TimerTask() {
					@Override
					public void run() {
						storage.put(key, value);
						adapter.wrote(contact, value);
						adapter.complete(put_times_out_before_completion);
					}
				},
				getOperationDelay());
	}
	
	private long getOperationDelay() {
		return (operation_delay +
		        (count_of_operation_with_additional_delay == operation_count
		        ? additional_delay : 0));
	}
	
	@Override
	protected void get(final byte[] key_bytes,
			           final DHTOperationAdapter adapter) {
		++operation_count;
		
		final HashWrapper key = new HashWrapper(key_bytes);
		send_answers_timer.schedule(
				new TimerTask() {
					@Override
					public void run() {
						boolean should_time_out = true;
						DHTTransportValue value = storage.get(key);
						if (value != null) {
							adapter.read(contact, value);
							should_time_out = false;
						}
						adapter.complete(should_time_out);
					}
				},
				getOperationDelay());
	}
}

// DHTVanishBAckend test class:

public class VuzeVanishBackendImplTest extends TestCase {	
	private TestDHTVanishBackend createBackendTest(
			int udp_timeout_min, int udp_timeout_max, long share_timeout,
			int num_tries) {
		VanishConfiguration conf = new VanishConfiguration();
		conf.setProperty(
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_MIN, udp_timeout_min);
		conf.setProperty(
				VuzeBackendDefaults.CONF_VUZE_UDP_TIMEOUT_MAX, udp_timeout_max);
		
		conf.setProperty(
				VuzeBackendDefaults.CONF_VANISH_BACKEND_VUZEIMPL_NUMTRIES,
				num_tries);
		
		TestDHTVanishBackend backend = null;
		try {
			backend = new TestDHTVanishBackend(conf);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to construct backend.");
		}
		
		try {
			backend.init();
		} catch (VanishBackendException e) {
			e.printStackTrace();
			fail("Failed to init backend.");
		}
		return backend;
	}
	
	public void testNoTimeouts() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);

		byte[][] locations = {
			"k1".getBytes(), "k2".getBytes(), "k3".getBytes() };
		byte[][] shares = {
			"s1".getBytes(), "s2".getBytes(), "s3".getBytes() };
		try {
			backend.pushShares(shares, locations);
		} catch (VanishBackendException e) {
			e.printStackTrace();
			fail("Failed to push shares.");
		}
		
		byte[][] result = null;
		try {
			result = backend.getShares(locations, 3);
		} catch(VanishException e) {
			e.printStackTrace();
			fail("Failed to get shares.");
		}
		
		assertNotNull(result);
		assertEquals(shares.length, result.length);
		for (int i = 0; i < shares.length; ++i) {
			assertTrue(Arrays.equals(shares[i], result[i]));
		}
		
		backend.stop();
	}
	
	public void testPutTimesOutBeforeBackendTimesOut() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);
		
		byte[][] locations = { "k1".getBytes() };
		byte[][] shares = { "s1".getBytes() };
		
		backend.put_times_out_before_completion = true;
		
		try {
			backend.pushShares(shares, locations);
			fail("Succeeded to push shares when it shouldn't have.");
		} catch (VanishBackendException e) {
			assertEquals(e.getMessage(), "Failed to push all shares.");
		}
		
		backend.stop();
	}
	
	public void testGetObtainsEnoughShares() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);
		
		byte[][] locations_put = { "k1".getBytes() };
		byte[][] shares_put = { "s1".getBytes() };
		byte[][] locations_get = { "k1".getBytes(), "k2".getBytes() };
		
		try {
			backend.pushShares(shares_put, locations_put);
		} catch (VanishBackendException e) {
			e.printStackTrace();
			fail("Failed to push shares.");
		}
		
		byte[][] result = null;
		try {
			result = backend.getShares(locations_get, 1);
		} catch(VanishException e) {
			e.printStackTrace();
			fail("Failed to get shares.");
		}
		
		assertNotNull(result);
		assertEquals(locations_get.length, result.length);
		for (int i = 0; i < shares_put.length; ++i) {
			assertNotNull(result[i]);
			assertTrue(Arrays.equals(shares_put[i], result[i]));
		}
		for (int i = shares_put.length; i < result.length; ++i) {
			assertNull(result[i]);
		}
		
		backend.stop();
	}
	
	public void testGetDoesntGetEnough() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);
		
		byte[][] locations_put = { "k1".getBytes() };
		byte[][] shares_put = { "s1".getBytes() };
		byte[][] locations_get = { "k1".getBytes(), "k2".getBytes() };
		
		try {
			backend.pushShares(shares_put, locations_put);
		} catch (VanishBackendException e) {
			e.printStackTrace();
			fail("Failed to push shares.");
		}
		
		byte[][] result = null;
		try {
			result = backend.getShares(locations_get, 2);
			fail("Got enough shares when it shouldn't have");
		} catch(TooManySharesLostVanishException e) {  // expected
			assertEquals("Got 1/2 shares", e.getMessage());
		} catch (Exception e) {
			e.printStackTrace();
			fail("Failed with wrong exception.");
		}
		
		assertNull(result);
		backend.stop();
	}
	
	public void testPutTimesOutBeforeAnyOperationCompletes() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);
		byte[][] locations_put = { "k1".getBytes() };
		byte[][] shares_put = { "s1".getBytes() };
		
		backend.operation_delay = 10000L;
		
		long start = System.currentTimeMillis();
		try {
			backend.pushShares(shares_put, locations_put);
			fail("Succeeded in pushing shares when it shouldn't have.");
		} catch (VanishBackendException e) {
			long end = System.currentTimeMillis();
			assertEquals("Failed to push all shares.", e.getMessage());
			assertTrue(end - start < 1000L);  // check it actually times out.
		}
		
		backend.stop();
	}
	
	public void testPutTimesOutBeforeSecondOperationCompletes() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);
		
		byte[][] locations_put = { "k1".getBytes(), "k2".getBytes() };
		byte[][] shares_put = { "s1".getBytes(), "s2".getBytes() };
		
		backend.additional_delay = 10000L;
		backend.count_of_operation_with_additional_delay = 2;
		
		long start = System.currentTimeMillis();
		try {
			backend.pushShares(shares_put, locations_put);
			fail("Succeeded in pushing shares when it shouldn't have.");
		} catch (VanishBackendException e) {
			long end = System.currentTimeMillis();
			assertEquals("Failed to push all shares.", e.getMessage());
			assertTrue(end - start < 1000L);
		}
		
		backend.stop();
	}
	
	public void testGetTimesOutBeforeSecondOperationCompletes() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);
		
		byte[][] locations_put = { "k1".getBytes(), "k2".getBytes() };
		byte[][] shares_put = { "s1".getBytes(), "s2".getBytes() };
		
		try {
			backend.pushShares(shares_put, locations_put);
		} catch (VanishBackendException e) {
			e.printStackTrace();
			fail("Failed to push shares.");
		}
		
		backend.additional_delay = 10000L;
		backend.count_of_operation_with_additional_delay = 4;
		
		long start = System.currentTimeMillis();
		try {
			backend.getShares(locations_put, 2);
			fail("Succeeded in pushing shares when it shouldn't have.");
		} catch (TooManySharesLostVanishException e) {
			long end = System.currentTimeMillis();
			assertEquals("Got 1/2 shares", e.getMessage());
			assertTrue(end - start < 1000L);  // check it actually times out.
		} catch (VanishBackendException e) {
			e.printStackTrace();
			fail("Failed to get shares.");
		}
		
		backend.stop();
	}
	
	public void testGetSucceedsAlthoughSecondOperationTimesOut() {
		TestDHTVanishBackend backend = createBackendTest(20, 20, 50, 1);
		
		byte[][] locations_put = { "k1".getBytes(), "k2".getBytes() };
		byte[][] shares_put = { "s1".getBytes(), "s2".getBytes() };
		
		try {
			backend.pushShares(shares_put, locations_put);
		} catch (VanishBackendException e) {
			e.printStackTrace();
			fail("Failed to push shares.");
		}
		
		backend.additional_delay = 10000L;
		backend.count_of_operation_with_additional_delay = 4;
		
		long start = System.currentTimeMillis();
		byte[][] result = null;
		try {
			result = backend.getShares(locations_put, 1);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to get shares.");
		}
		long end = System.currentTimeMillis();
		
		assertNotNull(result);
		assertEquals(2, result.length);
		assertNotNull(result[0]);
		assertTrue(Arrays.equals(shares_put[0], result[0]));
		assertNull(result[1]);
		
		assertTrue(end - start < 1000L);  // check it actually times out.
		
		backend.stop();
	}
	
	// Test the binary search for the UDP timeout:
	
	// TODO(roxana).
}


// Other helper functions:

class TestDHTTransportContact
implements DHTTransportContact {
	private String id;
	
	public TestDHTTransportContact(int id) {
		this.id = "" + id;
	}
	
	// @Override
	public void exportContact(DataOutputStream os)
	throws IOException, DHTTransportException { }

	// @Override
	public InetSocketAddress getAddress() { return null; }

	// @Override
	public long getClockSkew() { return 0; }

	// @Override
	public InetSocketAddress getExternalAddress() { return getAddress(); }

	// @Override
	public byte[] getID() { return id.getBytes(); }

	// @Override
	public int getInstanceID() { return id.hashCode(); }

	// @Override
	public int getMaxFailForLiveCount() { return 0; }

	// @Override
	public int getMaxFailForUnknownCount() { return 0; }

	// @Override
	public String getName() { return id; }

	// @Override
	public DHTNetworkPosition getNetworkPosition(byte position_type) {
		return null;
	}

	// @Override
	public DHTNetworkPosition[] getNetworkPositions() { return null; }

	// @Override
	public byte getProtocolVersion() { return 0; }

	// @Override
	public int getRandomID() { return id.hashCode(); }

	// @Override
	public DHTTransportFullStats getStats() { return null; }

	// @Override
	public String getString() { return id; }

	// @Override
	public DHTTransport getTransport() { return null; }

	// @Override
	public boolean isAlive(long timeout) { return true; }

	// @Override
	public boolean isValid() { return true; }

	// @Override
	public void remove() { }

	// @Override
	public void sendFindNode(DHTTransportReplyHandler handler, byte[] id) { }

	// @Override
	public void sendFindValue(DHTTransportReplyHandler handler, byte[] key,
			int max_values, byte flags) { }

	// @Override
	public void sendImmediatePing(DHTTransportReplyHandler handler,
			                      long timeout) { }

	// @Override
	public void sendKeyBlock(DHTTransportReplyHandler handler,
			byte[] key_block_request, byte[] key_block_signature) { }

	// @Override
	public void sendPing(DHTTransportReplyHandler handler) { }

	// @Override
	public void sendStats(DHTTransportReplyHandler handler) { }

	// @Override
	public void sendStore(DHTTransportReplyHandler handler, byte[][] keys,
			DHTTransportValue[][] value_sets, boolean immediate) { }

	// @Override
	public void setRandomID(int id) { }
}

class TestDHTTransportValue implements DHTTransportValue {
	private DHTTransportContact originator;
	private byte[] value;
	private boolean is_local;
	
	public TestDHTTransportValue(DHTTransportContact originator,
			 byte[] value, boolean is_local) {
		this.originator = originator;
		this.value = value;
		this.is_local = is_local;
	}
	
	// @Override
	public long getCreationTime() { return 0; }

	// @Override
	public int getFlags() { return 0; }

	// @Override
	public DHTTransportContact getOriginator() { return originator; }

	// @Override
	public String getString() { return "" + value; }

	// @Override
	public byte[] getValue() { return value; }

	// @Override
	public int getVersion() { return 0; }

	// @Override
	public boolean isLocal() { return is_local; }	
}


